@@ -25,11 +25,15 @@
 G_BEGIN_DECLS
 
 /* Open a dialog to create a new .grdc file. Return response id: OK:Connect, APPLY:Save, CANCEL */
-gint grdc_file_editor_create (GtkWindow *parent);
+gint grdc_file_editor_create (GtkWindow *parent, gboolean quit_when_close);
+gint grdc_file_editor_create_full (GtkWindow *parent, gboolean quit_when_close,
+    const gchar *server, const gchar *protocol);
 /* Open a dialog to edit an existing .grdc file. */
 gint grdc_file_editor_open (GtkWindow *parent, gchar *filename, gboolean quit_when_close);
 /* Open a dialog for a quick connection without .grdc file. */
 gint grdc_file_editor_open_temp (GtkWindow *parent, gboolean quit_when_close);
+gint grdc_file_editor_open_temp_full (GtkWindow *parent, gboolean quit_when_close,
+    const gchar *server, const gchar *protocol);
 
 G_END_DECLS
 
