@@ -30,7 +30,10 @@ G_BEGIN_DECLS
 
 enum
 {
-    GRDC_PREF_GROUPS_TAB = 1
+    GRDC_PREF_OPTIONS_TAB = 0,
+    GRDC_PREF_GROUPS_TAB = 1,
+    GRDC_PREF_RESOLUTIONS_TAB = 2,
+    GRDC_PREF_APPLET_TAB = 3
 };
 
 gboolean grdc_pref_dialog_run (GtkWindow *parent, gint default_tab);
